import os
import pytest

from unittest import mock

from socon.core.exceptions import HookNotFound

from socon_embedded.exceptions import ParserError
from socon_embedded.schema.task import Task


class TestDiscoverAction:

    def test_discover_custom_action_not_found(self):
        with pytest.raises(HookNotFound):
            Task(**{"name": "Test discover", "custom_action": {"test": "test"}})

    @mock.patch.dict(os.environ, {"SOCON_ACTIVE_PROJECT": "test_project"})
    def test_discover_custom_action_in_project(self):
        Task(**{"name": "Test discover", "custom_action": {"test": "test"}})



from socon_embedded.builder import Builder


class EchoBuilder(Builder):
    name = "echo"
    use_shell = True

    def get_executable(self) -> str:
        return "echo"

    def get_main_args(self, project_file, **variant_args) -> list[str]:
        return [project_file]



import os

from argparse import ArgumentParser
from collections.abc import MutableMapping
from typing import Any, Dict, Optional, Tuple

from socon_embedded.utils.loader import from_json_or_yaml, load_from_file
from socon_embedded.utils.converter import to_text
from socon_embedded.utils.parser import parse_key_value
from socon.conf import settings

from socon.core.management.base import CommandError, Config, ProjectCommand
from socon.core.management.subcommand import Subcommand
from socon.core.registry.config import ProjectConfig


class BuildCommand(Subcommand):
    help = "Build a project application"
    subcommand_manager = "build_manager"


class BuildCommandInterface(ProjectCommand, abstract=True):
    manager = "build_manager"

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--app", help="The name of the application to build", action="append"
        )
        parser.add_argument(
            "--group",
            help="The group name of the applications to build",
            action="append",
        )
        parser.add_argument(
            "--builder", help="The name of the builder", action="append"
        )
        parser.add_argument(
            "--base-dir",
            help=("Base build directory for the application that needs to be build"),
        )
        parser.add_argument(
            "--filter",
            help=(
                "Set filter to build only specific application. "
                'Example: --filter "name=myapp", --filter "group__in=[\'xx\']"'
            ),
            action="append",
        )
        parser.add_argument(
            "--exclude",
            help=(
                "Set filter to exclude specific application from being built. "
                'Example: --exclude "name=myapp", --exclude "group__in=[\'xx\']"'
            ),
            action="append",
        )
        parser.add_argument(
            "--variant-args",
            help=(
                "Filter a build variant args configiration. "
                "Like a mode variant configuration. "
                'Example: --variant-args "mode=release"'
            ),
            action="append",
        )
        parser.add_argument(
            "--extras-vars",
            help=(
                "Set variables as key=value or YAML/JSON, if filename prepend with @. "
                'Example: --set-vars "fruit=apple", --set-vars "@file.yml"'
            ),
            action="append",
        )
        parser.add_argument(
            "--eoe",
            "--exit-on-error",
            help="Exit if there is an error",
            action="store_true",
        )
        parser.add_argument(
            "--wae", "--warning-as-error", help="Treat a warning as an error"
        )
        parser.add_argument(
            "--artifact-dir",
            help=("Path to the artifact folder that will store " "the build artifacts"),
        )
        self.add_build_arguments(parser)

    def add_build_arguments(self, parser: ArgumentParser) -> None:
        """Add more arguments to the base build arguments"""
        pass

    def handle(self, config: Config, project_config: ProjectConfig) -> str:
        # Transform each of the following command line args as a dictionary
        filters = self.load_template_args(config.getoption("filter", []))
        excludes = self.load_template_args(config.getoption("excludes", []))
        extras_vars = self.load_template_args(config.getoption("extras_vars", []))
        variant_args = self.load_template_args(config.getoption("variant_args", []))

        # Merge the extras filters with the default one passed on the command line
        filters_opt = self._create_filter(
            ("name__in", config.getoption("app")),
            ("group__in", config.getoption("group")),
            ("builders__name__in", config.getoption("builder")),
        )
        filters = filters | filters_opt

        # Create the variant config filter
        variant_args_filters = {}
        for var_config, value in variant_args.items():
            variant_args_filters.update({f"variant_args__{var_config}": value})

        # Create a context based on the extras vars and the base directory. This
        # will be mainly use to resolve any further jinja template file
        base_dir = config.getoption("base_dir")
        context = extras_vars
        if base_dir is not None:
            context.update({"base_dir": base_dir})

        # Get exit condition and pass them to the handle_build method.
        eoe = config.getoption("eoe")
        wae = config.getoption("wae")

        # Get the output directory if any
        artifact_dir = config.getoption("artifact_dir") or getattr(
            settings, "BUILD_ARTIFACT_PATH"
        )

        # Load every variable that needs to be export in the project config
        env_variables = project_config.get_setting(
            "BUILD_ENVIRONMENT_VARIABLE", skip=True, default=[]
        )
        for env, value in env_variables:
            if os.environ.get(env) is not None:
                os.environ[env] += os.pathsep + value
            else:
                os.environ[env] = value

        # Pass all the above to a method used by each subcommands
        self.handle_build(
            config=config,
            project_config=project_config,
            filters=filters,
            excludes=excludes,
            variant_args_filters=variant_args_filters,
            exit_on_error=eoe,
            context=context,
            warning_as_error=wae,
            artifact_dir=artifact_dir,
        )

    def handle_build(
        self,
        config: Config,
        project_config: ProjectConfig,
        filters: dict = {},
        excludes: dict = {},
        variant_args_filters: dict = {},
        context: dict = {},
        exit_on_error: bool = False,
        warning_as_error: bool = False,
        artifact_dir: str = None,
    ) -> str:
        raise NotImplementedError(
            "Subclass of '{}' must implement handle_build(...) method".format(
                self.__class__.__name__
            )
        )

    @staticmethod
    def load_template_args(variables: list[Any] = tuple()) -> Dict[str, str]:
        """Load vars passed in command line arguments"""
        loaded_vars = {}

        for var_opt in variables:
            data = None
            var_opt = to_text(var_opt)
            if var_opt is None or not var_opt:
                continue
            if var_opt.startswith("@"):
                # Argument is a YAML file (JSON is a subset of YAML)
                data = load_from_file(var_opt[1:])
            elif var_opt[0] in ["/", "."]:
                raise CommandError(
                    f"Please prepend set-vars filename '{var_opt}' with '@'"
                )
            elif var_opt[0] in ["[", "{"]:
                # Arguments as YAML
                data = from_json_or_yaml(var_opt)
            else:
                # Arguments as Key-value
                data = parse_key_value(var_opt)

            if isinstance(data, MutableMapping):
                loaded_vars = loaded_vars | data
            else:
                raise CommandError(
                    f"Invalid extra vars data supplied. '{var_opt}' could "
                    "not be made into a dictionary"
                )

        return loaded_vars

    @staticmethod
    def _create_filter(*filters: Tuple[str, Optional[dict]]) -> dict:
        output = {}
        for key, value in filters:
            if value is not None:
                output[key] = value
        return output



from __future__ import annotations

from typing import List, Literal, Optional, Type, Any

from socon.core.registry import projects

from socon_embedded.exceptions import ParserError
from socon_embedded.managers import get_action_manager
from socon_embedded.schema.base import Base, Nameable, get_field_names

from pydantic import BaseModel, Field, model_validator


class Task(Base, Nameable):
    """Call a module (action) with specific arguments and other parameters"""

    action: Any
    args: dict = {}

    retries: Optional[int] = 0
    timeout: Optional[int] = None

    @model_validator(mode="before")
    @classmethod
    def preprocess_action(cls, values):
        action = None
        args = {}

        if "action" in values:
            action = values["action"]
        if "args" in values:
            args = values["args"]

        # If the application registry filters some of the apps,
        # the validation will need to be retriggered and the action
        # will already be an instance. To avoid that, we check if we already have an action instance and return it as his.
        if isinstance(action, object) and action is not None:
            return values

        # filter out task attributes so we're only querying unrecognized keys as actions/modules
        non_task_values = dict(
            (k, v) for k, v in values.items() if (k not in get_field_names(Task))
        )

        # walk the filtered input dictionary to see if we recognize a module name
        action_manager = get_action_manager()
        for item, value in non_task_values.items():
            is_action_candidate = False
            if item in action_manager.get_hooks_name():
                is_action_candidate = True

            if is_action_candidate:
                # finding more than one module name is a problem
                if action is not None:
                    raise ParserError(
                        "conflicting action statements: {}, {}".format(action, item)
                    )

                action = item
                args = value

        # if we didn't see any module in the task at all, it's not a task really
        if action is None:
            if (
                non_task_values
            ):  # there was one non-task action, but we couldn't find it
                bad_action = list(non_task_values.keys())[0]
                raise ParserError(
                    "couldn't resolve module/action '{0}'. This often indicates a "
                    "misspelling, missing collection, or "
                    "incorrect module path.".format(bad_action)
                )
            else:
                raise ParserError("no module/action detected in task.")

        try:
            project_config = projects.get_project_config_by_env()
        except LookupError:
            project_config = None
        action = action_manager.search_hook_impl(action, project_config)

        # Set the action if no previous error
        values["action"] = action
        values["args"] = args
        return values

    @model_validator(mode="after")
    def post_process_action(self):
        """Validate the action schema"""
        if isinstance(self.action, type):
            self.action = self.action(self)
        return self


class Taskable(BaseModel):
    """
    Make a schema taskable.
    Give the ability to have pre_tasks, post_tasks and tasks
    """

    # Tasks list attribute
    tasks: List[Task] = Field(default_factory=list)
    post_tasks: List[Task] = Field(default_factory=list)

    def merge_tasks(
        self, y: Type[Taskable], keep: bool, tasks_type: Literal["tasks", "post_tasks"]
    ) -> Type[Taskable]:
        """Merge y tasks into x tasks"""
        if keep is False:
            setattr(self, tasks_type, getattr(y, tasks_type))
        else:
            for task in getattr(y, tasks_type):
                if task not in getattr(self, tasks_type):
                    getattr(self, tasks_type).append(task)
        return self



from socon.core.management import call_command


class TestFromFileCommand:
    def test_build_from_file(self, tmpdir, datafix_dir):
        tmp = tmpdir.mkdir("artifact")
        call_command(
            "build",
            "fromfile",
            "--file",
            f"{datafix_dir}/simple_app_config.yml",
            "--project",
            "test_project",
            "--artifact-dir",
            tmp,
        )



from __future__ import annotations
import itertools

from typing import Dict, List, Optional, Union
from socon_embedded.builder import BuildInfo

from socon_embedded.exceptions import YamlFormatError, YamlParserError
from socon_embedded.managers import get_builder_manager
from socon_embedded.schema.task import Taskable
from socon_embedded.schema.base import Base, Nameable, get_field_names

from pydantic import BaseModel, field_validator, model_validator

from datalookup import Dataset


class AppRegistry(Base, Nameable, Taskable):
    name: Optional[str]
    vars: Optional[dict] = {}
    apps: List[AppConfig] = []

    def filter(self, filters: dict = {}, excludes: dict = {}) -> AppRegistry:
        """
        Filter using datalookup library, Return a new registry
        with the filtered apps
        """
        if filters == {} and excludes == {}:
            return self

        dataset = Dataset(self.model_dump()["apps"])

        # Filter and excludes apps
        apps = dataset.on_cascade().filter(**filters)
        if excludes:
            apps = apps.on_cascade().exclude(**excludes)

        # Re-create the AppConfig object for each filtered apps
        filtered_apps = []
        for app in apps:
            filtered_apps.append(AppConfig(**app.values()))

        # Return a new registry with the filtered apps
        return AppRegistry(**self.model_dump(exclude="apps"), apps=filtered_apps)

    def _get_app(self, name: str) -> Optional[AppConfig]:
        """Get an application from the registry"""
        for app in self.apps:
            if name == app.name:
                return app
        return None

    def add_application(self, name: str, group: list = []) -> AppConfig:
        app = self._get_app(name)
        if app is None:
            app = AppConfig(name=name, group=group)
            self.apps.append(app)
        else:
            app.add_group(group)
        return app

    def add_vars(self, **variables):
        """Add global variable to the registry"""
        self.vars |= variables


class AppConfig(Base, Nameable, Taskable):
    builders: List[AppBuilder] = []
    group: Optional[Union[str, int, list]] = None
    variants: Optional[List[Variant]] = []

    @field_validator("group", mode="before")
    @classmethod
    def convert_group_to_list(cls, v: Union[str, int, list]):
        if isinstance(v, (str, int)):
            return [v]
        return v

    @model_validator(mode="after")
    def validate_builder_exist(self):
        # Find common builders and project builders if the project define
        builder_manager = get_builder_manager()

        not_existing_builders = []
        for builder in self.builders:
            if builder.name not in builder_manager.get_hooks_name():
                not_existing_builders.append(builder.name)

        if not_existing_builders:
            raise YamlParserError(
                "Following builder(s) does not exist(s):\n" f"{not_existing_builders}"
            )

        return self

    @model_validator(mode="after")
    def validate_builder_ref(self):
        builders = [builder.name for builder in self.builders]
        for variant in self.variants:
            builder_refs = [ref for builder in variant.builders for ref in builder.ref]
            for ref in builder_refs:
                if ref not in builders:
                    raise YamlFormatError(
                        f"Reference to '{ref}' builder does not exist.\n"
                        f"Please check following variant:\n{variant}"
                    )
        return self

    @staticmethod
    def _get_build_configs(
        app: str, builder: AppBuilder, filters: dict = {}
    ) -> list[BuildConfig]:
        build_configs = []

        # If the user specified the configs entry, we need to create
        # multiple build configuration
        if builder.variant_args:
            keys, values = zip(*builder.variant_args.items())
            permutations_dicts = [
                dict(zip(keys, v)) for v in itertools.product(*values)
            ]
            for variant in permutations_dicts:
                config_builder = AppBuilder(
                    **builder.model_dump(exclude="variant_args"), variant_args=variant
                )
                app_dataset = Dataset(config_builder.model_dump())
                if len(app_dataset.filter(**filters)) != 0:
                    build_configs.append(BuildConfig(app=app, builder=config_builder))
        else:
            build_configs.append(BuildConfig(app=app, builder=builder))

        return build_configs

    def get_build_configs(self, filters: dict = {}) -> List[BuildConfig]:
        build_configs = []
        builders_ref: Dict[str, AppBuilder] = {}

        for builder in self.builders:
            # Save the builder reference for the variant
            if builder.name not in builders_ref:
                builders_ref[builder.name] = builder

            # If the user specified the configs entry, we need to create
            # multiple build configuration
            build_configs.extend(self._get_build_configs(self.name, builder, filters))

        for variant in self.variants:
            app = self.name + f".{variant.name}"
            for vbuilder in variant.builders:
                for ref in vbuilder.ref:
                    builder_ref = builders_ref[ref]
                    builder = builder_ref.merge_variant_builder(vbuilder)
                    build_configs.extend(self._get_build_configs(app, builder, filters))

        return build_configs

    def _get_builder(self, name: str) -> Optional[AppBuilder]:
        for builder in self.builders:
            if name == builder.name:
                return builder
        return None

    def add_group(self, value: str) -> None:
        """Add a value to the application config group"""
        if not self.group:
            self.group = [value]
        elif value not in self.group:
            self.group.append(value)

    def add_builder(self, name: str, project_file: str, extras_options: dict = None):
        """Add a builder to the application"""
        builder = self._get_builder(name)
        if builder is None:
            builder = AppBuilder(
                name=name, project_file=project_file, variant_args=extras_options
            )
            self.builders.append(builder)
        return builder


class BuildConfig(Base, Taskable):
    app: str
    builder: AppBuilder

    def get_buildinfo(self) -> dict:
        return {
            "app": self.app,
            **self.builder.model_dump(
                by_alias=True,
                exclude=[*get_field_names(Taskable), *get_field_names(Nameable)],
            ),
        }

    def create_buildinfo(self) -> BuildInfo:
        build_info = self.get_buildinfo()
        build_info.pop("raw_args")
        return BuildInfo(builder=self.builder.name, **build_info)


class Builder(Taskable):
    variant_args: Optional[Dict[str, Union[str, list]]] = {}
    raw_args: Optional[list] = []


class AppBuilder(Base, Builder, Nameable):
    project_file: str

    def merge_variant_builder(self, other: VariantBuilder) -> AppBuilder:
        """Merge a builder with a variant builder"""
        builder = self.model_copy(deep=True)

        # Merge tasks from other to the current builder
        builder.merge_tasks(other, other.keep_tasks, "tasks")
        builder.merge_tasks(other, other.keep_post_tasks, "post_tasks")

        # Merge all remaining fields
        builder_dict = builder.model_dump() | other.model_dump(
            exclude=[*get_field_names(Taskable), *get_field_names(VariantBuilderField)]
        )

        return AppBuilder(**builder_dict)


class VariantBuilderField(BaseModel):
    ref: Union[str, list]
    keep_tasks: Optional[bool] = False
    keep_post_tasks: Optional[bool] = False

    @field_validator("ref", mode="before")
    @classmethod
    def convert_ref_to_list(cls, v: Union[str, list]):
        if isinstance(v, str):
            return [v]
        return v


class VariantBuilder(Builder, VariantBuilderField):
    pass


class Variant(Base, Nameable, Taskable):
    group: Optional[Union[str, int, list]] = None
    builders: Optional[List[VariantBuilder]] = []



import socon

from socon.conf import global_settings


def pytest_configure(config):
    from socon.conf import settings

    settings.configure(
        default_settings=global_settings,
        INSTALLED_PLUGINS=("socon_embedded",),
        INSTALLED_PROJECTS=("projects.test_project",),
    )

    socon.setup()



import json
import traceback

from typing import Optional
from collections import OrderedDict

from socon_embedded.executor.task_result import TaskResult
from socon_embedded.schema.task import Task
from socon_embedded.utils.converter import to_text

from socon.utils.terminal import terminal


class TaskTimeoutError(BaseException):
    pass


def task_timeout(signum, frame):
    raise TaskTimeoutError


class TaskPlayer:

    def __init__(self) -> None:
        self._played_tasks: list[TaskExecutor] = []
        self._terminal = terminal

    def run(self, tasks: list[Task]) -> None:
        for task in tasks:
            executor = TaskExecutor(task, {})
            self._task_on_start(task)
            result = executor.run()
            tr = TaskResult(task, result)
            if tr.is_failed():
                self._task_on_failed(tr)
                executor.cleanup()
                for played_exec in self._played_tasks:
                    played_exec.cleanup()
                return
            self._task_on_ok(tr)
            self._played_tasks.append(executor)
        if len(self._played_tasks) != 0:
            self._terminal.line()

    def cleanup(self) -> None:
        for executor in self._played_tasks:
            executor.cleanup()

    def _task_on_start(self, task: Task):
        if self._terminal._current_line != "":
            self._terminal.line()
        self._terminal.line(f"TASK [{task.name}]")
        self._terminal.sep("*")

    def _task_on_failed(self, result: TaskResult):
        msg = "failed: => {}".format(self._dump_results(result._result))
        self._terminal.line(msg, fg="red")

    def _task_on_ok(self, result: TaskResult):
        self._terminal.line("succesfull")

    def _dump_results(
        self,
        result: dict,
        indent: Optional[int] = 4,
        sort_keys: bool = False,
    ) -> str:
        try:
            return json.dumps(
                result, indent=indent, ensure_ascii=False, sort_keys=sort_keys
            )
        except TypeError:
            return json.dumps(
                OrderedDict(sorted(result.items(), key=to_text)),
                indent=indent,
                ensure_ascii=False,
                sort_keys=False,
            )


class TaskExecutor:

    def __init__(self, task: Task, job_vars: dict = {}) -> None:
        self._task = task
        self._job_vars = job_vars
        self._terminal = terminal

    def run(self) -> dict:
        try:
            res = self._execute()
            return res
        except Exception as e:
            return dict(
                failed=True,
                msg=f"Unexpected failure during module execution: {e}",
                exception=to_text(traceback.format_exc()),
                stdout="",
            )

    def _execute(self, variables: Optional[dict] = None) -> dict:
        if variables is None:
            variables = self._job_vars
            vars_copy = variables.copy()

        retries = 1  # includes the default actual run + retries set by user/default
        if self._task.retries is not None:
            retries += max(0, self._task.retries)

        result = {}
        for attempt in range(1, retries + 1):
            try:
                result = self._task.action.run(tasks_vars=vars_copy)
            except Exception as e:
                self._task.action.cleanup()
                raise e

            # set the failed property if it was missing.
            if "failed" not in result:
                result["failed"] = False

            # Make attempts and retries available early to allow their use in changed/failed_when
            if retries > 1:
                result["attempts"] = attempt
        else:
            if retries > 1:
                # we ran out of attempts, so mark the result as failed
                result["attempts"] = retries - 1
                result["failed"] = True

        return result

    def cleanup(self):
        self._task.action.cleanup()

    def _task_on_retry(self, result: dict):
        msg = "FAILED - RETRYING:  ({} retries left).".format(
            self._task.name, result["retries"] - result["attempts"]
        )
        self._terminal.line(msg)



import os
from pathlib import Path
import shutil

from typing import Any, Dict, Union

from socon_embedded.builder.result import BuildResult, Result, Status
from socon_embedded.managers import BuilderManager
from socon_embedded.schema.apps import AppRegistry, BuildConfig, AppConfig
from socon_embedded.builder import BuildInfo, Builder
from socon_embedded.executor.task_executor import TaskPlayer

from socon.core.registry import projects
from socon.core.registry.config import ProjectConfig
from socon.utils.terminal import terminal

from junitparser import JUnitXml, TestCase, TestSuite, Failure, Skipped


class AppRegistryExecutor:

    def __init__(
        self,
        app_registry: AppRegistry,
        builder_manager: BuilderManager,
        project_config: ProjectConfig = None,
    ) -> None:
        self._app_registry = app_registry
        self._builder_manager = builder_manager
        self._project_config = project_config

        # Load the project config if none are given
        if self._project_config is None:
            try:
                self._project_config = projects.get_project_config_by_env()
            except LookupError:
                pass

        # Cache for used builder when building the application in the registry
        self._cached_builders: Dict[str, Builder] = {}

        # Cache for applications results
        self._build_results: list[BuildResult] = []

    def _clear_cache(self):
        self._cached_builders = {}
        self._build_results = []

    @classmethod
    def from_file(cls, file: Union[str, os.PathLike], context: dict = {}, **kwargs):
        """Create an AppRegistry from a yaml/json file"""
        app_registry = AppRegistry.load(file, context)
        return cls(app_registry, **kwargs)

    def build(
        self,
        filters: Dict[str, Any] = {},
        excludes: Dict[str, Any] = {},
        variant_args_filters: Dict[str, Any] = {},
        output_dir: Union[str, os.PathLike] = None,
        exit_on_error: bool = False,
        warning_as_error: bool = False,
    ) -> AppRegistry:
        """Build the application in each registries based on the given filters"""
        self._clear_cache()

        # Re-define output_dir if not given
        output_dir = self._get_output_dir(output_dir)

        # Create a build section
        terminal.sep(
            "-", f"Application registry: {self._app_registry.name}", newline="both"
        )

        # Filter the application and return a AppRegistry with only the application
        # that we want to build
        reg = self._app_registry.filter(filters, excludes)

        # Raise a LookupError if we don't find any build configuration
        if not reg.apps:
            raise LookupError(
                "Nothing to build. You should check if you have registered your\n"
                "applications or if your filters are correct."
            )

        self.pre_build(reg)

        # Stop building in case exit_on_error is True and an issue was found.
        # This flag allow to still create a report with the rest of the application
        # set as skipped
        stop_building = False

        # Run the general tasks in priority
        task_player = TaskPlayer()
        task_player.run(reg.tasks)

        # Build each build configuration
        for app_config in reg.apps:

            # Run the Application config tasks
            task_player.run(app_config.tasks)

            self.post_process_app_config(app_config)

            # Get all build configuration for each application
            build_configs = app_config.get_build_configs(variant_args_filters)

            for build_config in build_configs:

                # Run the build config tasks
                task_player.run(build_config.tasks)

                # Create a build info object
                build_info = build_config.create_buildinfo()

                # If stop building is True, we still create result with skipped
                # status and the build info
                if stop_building is True:
                    result = BuildResult(
                        Result(Status.SKIPPED, "Skipped due to previous error")
                    )
                    result.build_info = build_info
                    self._build_results.append(result)
                    continue

                # Create the artifact directory
                artifact_path = self._create_artifact_directory(build_info, output_dir)

                self.pre_build_config(build_config)

                # Build the application
                builder = self._get_builder(build_config.builder.name)
                result = builder.build(
                    **build_config.get_buildinfo(),
                    output_file=Path(artifact_path, f"{build_info.app}.log"),
                    warning_as_error=warning_as_error,
                    variables=self._app_registry.vars,
                )

                self.post_build_config(build_config, result)

                # Save the fail result of the current apps
                self._build_results.append(result)

                # In case we need to exit on error. We need to stop the build
                # of all application but still put the next application that
                # were not build as skipped for the junit report.
                if result.is_fail and exit_on_error is True:
                    stop_building = True

                # Run the post build configs only if we decided not to stop
                # building.
                task_player.run(build_config.post_tasks)

            # Run the post application config tasks
            task_player.run(app_config.post_tasks)

        task_player.run(reg.post_tasks)

        # Call the post_build method for the user
        self.post_build(reg, output_dir)

        # Clean all the tasks at the end
        task_player.cleanup()

        return reg

    def pre_build(self, registry: AppRegistry):
        """Pre build method with the filtered application registry"""

    def post_build(self, registry: AppRegistry, output_dir: str):
        """Post build method with the filtered application registry"""

    def pre_build_config(self, build_config: BuildConfig):
        """Pre build config method with the configuration of the build"""

    def post_build_config(self, build_config: BuildConfig, result: Result):
        """Post build config method started after the build"""

    def post_process_app_config(self, app_config: AppConfig):
        """Post processing the application config"""

    def _get_output_dir(self, output_dir: str = None) -> str:
        output_dir = output_dir if output_dir else os.getcwd()
        return Path(output_dir, self._app_registry.name)

    def _create_artifact_directory(
        self, build_info: BuildInfo, output_dir: str, clean: bool = False
    ):
        """
        Save the output in an artifact directory at the current location or
        inside the specified output directory.
        If clean = True, the directory will be cleaned
        """
        artifact_path = list(
            filter(
                None,
                [
                    output_dir,
                    build_info.builder,
                    build_info.app,
                    "_".join(sorted(build_info.variant_args.values())),
                ],
            )
        )
        artifact_path = Path().joinpath(*artifact_path)

        # In case the result folder does not exist, we create it. In case the
        # clean options is set, we remove and recreate the folder.
        if artifact_path.is_dir():
            if clean:
                shutil.rmtree(artifact_path)
                os.mkdir(artifact_path)
        else:
            os.makedirs(artifact_path, exist_ok=True)

        return artifact_path

    def create_report(
        self, output_file: str, output_dir: str = None, add_skipped_apps: bool = True
    ) -> None:
        """Create junit report from buidled and skipped apps"""
        output_dir = self._get_output_dir(output_dir)
        junit_file = Path(output_dir) / output_file

        # Create the Junit object
        junit = JUnitXml()

        # If there is no builded application, we return an empty junitxml file
        if not self._build_results:
            junit.write(str(junit_file), pretty=True)
            return

        # Iterate over all build results
        testsuite = TestSuite(self._app_registry.name)
        for result in self._build_results:
            testcase = TestCase(result.build_info.get_case_name())
            testcase.time = result.execution_time

            # Assign a Failure or Skipped result to the testcase
            if result.is_fail:
                testcase.result = [Failure(result.result.text)]
            elif result.is_skipped:
                testcase.result = [Skipped(result.result.text)]

            # Do not add testcase that are skipped if add_skipped_apps is False
            if result.is_skipped and add_skipped_apps is False:
                continue

            testsuite.add_testcase(testcase)

        # Add the testsuite to the junit collection
        junit.add_testsuite(testsuite)

        # Create the report
        junit.write(str(junit_file), pretty=True)

        # Return the junit report in case someone needs to use it
        return junit

    def _get_builder(self, name: str) -> Builder:
        """Get the builder in cache or via the manager"""
        if name in self._cached_builders:
            builder = self._cached_builders.get(name)
        else:
            builder_klass = self._builder_manager.search_hook_impl(
                name, self._project_config
            )
            builder = builder_klass()
            self._cached_builders[name] = builder

        return builder



import copy
import json
import logging
import os

from pathlib import Path
from typing import Union

from yaml import YAMLError, safe_load

from socon_embedded.exceptions import YamlParserError

logger = logging.getLogger(__name__)


def from_json_or_yaml(data, show_content=True):
    """
    Creates a python datastructure from the given data, which can be either
    a JSON or YAML string.
    """
    new_data = None

    try:
        new_data = json.loads(data)
    except json.JSONDecodeError as json_exc:
        try:
            new_data = safe_load(data)
        except YAMLError as yaml_exc:
            raise YamlParserError(
                "We were unable to read either as JSON nor YAML, these are the "
                "errors we got from each:\n"
                f"JSON: {json_exc}\n\n"
                f"YAML: {yaml_exc}\n\n"
                f"Content of the file:\n{new_data}"
                if show_content is True
                else ""
            )

    return new_data


def load_from_file(file: Union[str, os.PathLike], show_content: bool = True):
    """Loads data from a file, which can contain either JSON or YAML."""

    # Transform the file as Path and resolve it. Resolving the file might
    # trigger FileNotFoundError
    file = Path(file).expanduser().resolve(True)

    logger.debug("Loading data from {}".format(file))

    try:
        with open(file, "r") as fp:
            file_data = fp.read()
    except (IOError, OSError) as e:
        raise YamlParserError(
            f"An error occurred while trying to read the file: {file}"
        ) from e

    parsed_data = from_json_or_yaml(data=file_data, show_content=show_content)

    return copy.deepcopy(parsed_data)



from socon_embedded.action import ActionBase
from pydantic import BaseModel


class CustomSchema(BaseModel):
    test: str


class CustomAction(ActionBase):
    name = "custom_action"
    schema = CustomSchema

    def run(self, tasks_vars=None):
        result = super().run(tasks_vars)
        result["discoverd"] = True
        return result



import pytest

from socon_embedded.utils.parser import parse_key_value, split_args


# A list of arguments that can be passed as a command line arguments
SPLIT_DATA = (
    (None, [], {}),
    ("", [], {}),
    ("a=b", ["a=b"], {"a": "b"}),
    ("a=['a', 'b']", ["a=['a', 'b']"], {"a": ["a", "b"]}),
    ("a=\"['a', 'b']\"", ["a=\"['a', 'b']\""], {"a": ["a", "b"]}),
    ('a="foo bar"', ['a="foo bar"'], {"a": "foo bar"}),
    ('a=b c="foo bar"', ["a=b", 'c="foo bar"'], {"a": "b", "c": "foo bar"}),
    (
        'a="echo \\"hello world\\"" b=bar',
        ['a="echo \\"hello world\\""', "b=bar"],
        {"a": 'echo "hello world"', "b": "bar"},
    ),
    ('a="nest\'ed"', ['a="nest\'ed"'], {"a": "nest'ed"}),
    ('a="multi\nline"', ['a="multi\nline"'], {"a": "multi\nline"}),
    ('a="blank\n\nline"', ['a="blank\n\nline"'], {"a": "blank\n\nline"}),
    ('a="blank\n\n\nlines"', ['a="blank\n\n\nlines"'], {"a": "blank\n\n\nlines"}),
    (
        'a="a long\nmessage\\\nabout a thing\n"',
        ['a="a long\nmessage\\\nabout a thing\n"'],
        {"a": "a long\nmessage\\\nabout a thing\n"},
    ),
    (
        'a="multiline\nmessage1\\\n" b="multiline\nmessage2\\\n"',
        ['a="multiline\nmessage1\\\n"', 'b="multiline\nmessage2\\\n"'],
        {"a": "multiline\nmessage1\\\n", "b": "multiline\nmessage2\\\n"},
    ),
    ('a="café eñyei"', ['a="café eñyei"'], {"a": "café eñyei"}),
    ("a=café b=eñyei", ["a=café", "b=eñyei"], {"a": "café", "b": "eñyei"}),
)

# Data for each tests. First input for both the data
SPLIT_ARGS = tuple((test[0], test[1]) for test in SPLIT_DATA)
PARSE_KV = tuple((test[0], test[2]) for test in SPLIT_DATA)


@pytest.mark.parametrize(
    "args, expected", SPLIT_ARGS, ids=[str(arg[0]) for arg in SPLIT_ARGS]
)
def test_split_args(args, expected):
    """Check that mulitple command line arguments are split into a list of args"""
    assert split_args(args) == expected


@pytest.mark.parametrize(
    "args, expected", PARSE_KV, ids=[str(arg[0]) for arg in PARSE_KV]
)
def test_parse_kv(args, expected):
    """
    Parse key value command line arguments and check that we get
    a correct python object
    """
    assert parse_key_value(args) == expected



import os
from typing import Type, Union
from pydantic import BaseModel

from socon_embedded.utils.jinja import jinja_resolve
from socon_embedded.utils.loader import from_json_or_yaml, load_from_file


def get_field_names(model: Type[BaseModel]) -> list[str]:
    return list(model.model_fields.keys())


class Base(BaseModel):

    @classmethod
    def load(cls, file: Union[str, os.PathLike], context: dict = {}):
        """Create a schema from a yaml/json file"""
        if context:
            content = jinja_resolve(file, context)
            data = from_json_or_yaml(content)
        else:
            data = load_from_file(file)
        return cls(**data)


class Nameable(BaseModel):
    name: str



"""socon-embedded exceptions"""


class YamlParserError(Exception):
    """something was detected early that is wrong about app config gile"""


class YamlFormatError(Exception):
    """Raised when the application registry yaml format is wrong"""


class ApplicationRegistryNotReady(Exception):
    """Raised when the application registry is not ready"""


class ParserError(Exception):
    """Raised when we cannot passe or load a YAML/Json file"""


class TaskExecutionError(Exception):
    """Raised when a task failed for any exception"""



from __future__ import annotations

import logging
import os
from pathlib import Path
import subprocess
import sys
import time

from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, Mapping, Optional, Sequence, Tuple, Union

from socon_embedded.builder.parser import DefaultParser
from socon_embedded.builder.result import Result, Status
from socon_embedded.utils.converter import safe_decode
from socon_embedded.builder.result import BuildResult

from socon.core.manager import Hook
from socon.utils.terminal import terminal


logger = logging.getLogger(__name__)


class BuildCommandNotFound(Exception):
    def __init__(self, command: list[str], cause: Exception) -> None:
        self.command = command
        cause = "{}('{}')".format(type(cause).__name__, safe_decode(str(cause)))
        self._cmd = safe_decode(self.command[0])
        self._cmdline = " ".join(safe_decode(i) for i in self.command)
        self._cause = cause and f" due to: {cause}" or "!"

    def __str__(self) -> str:
        return ("Cmd('{}') not found{}" + "\n  cmdline: {}").format(
            self._cmd, self._cause, self._cmdline
        )


@dataclass
class BuildInfo:
    """Store building information"""

    app: str
    builder: str
    project_file: str
    variant_args: dict = field(default_factory=dict)
    cmdline: list = field(default_factory=list)

    def __str__(self) -> str:
        output = [
            f"Application: {self.app}",
            f"Input file: {self.project_file}",
            f"Builder: {self.builder}",
        ]
        variant_args = self.variant_args
        if variant_args:
            for key, value in variant_args.items():
                output.append(f"{key.title()}: {value}")
        return "\n".join(output)

    def get_case_name(self) -> str:
        """Get the testcase name that will be used in the junit report"""
        return " - ".join([self.app, self.builder, *self.variant_args.values()])


class Builder(Hook, abstract=True):
    """Base class for toolchain or compiler"""

    manager = "builder"

    # Default options defined by the user
    persistant_options: list[str] = []

    # If True, a shell will be used when executing git commands.
    use_shell = False

    def __init__(
        self,
        name: Optional[str] = None,
        executable: Optional[Union[str, os.PathLike]] = None,
    ) -> None:
        self.name = name or getattr(self, "name", self.__class__.__name__)
        self.executable = executable or self.get_executable()
        if not self.executable:
            raise AttributeError(
                "'executable' attribute is required by '{}'".format(self.name)
            )

        # User can define the parser as a class attribute if needed
        if not hasattr(self, "parser"):
            self.parser = DefaultParser

    def get_executable(self) -> str:
        """Get the builder executable"""
        return getattr(self, "executable")

    def build(
        self,
        app: str,
        project_file: Union[str, os.PathLike],
        variant_args: dict = {},
        raw_args: list[str] = [],
        warning_as_error: bool = False,
        output_file: Union[str, os.PathLike] = None,
        clean: bool = False,
        variables: dict = {},
        **kwargs: Any,
    ) -> BuildResult:
        """
        Compile an application using the input path, the specified mode and arguments
        """
        main_args = self.get_main_args(project_file, **variant_args)
        if isinstance(main_args, str):
            main_args = main_args.split()

        # Raise an error if main_args is None
        if main_args is None:
            raise RuntimeError(
                "'get_main_args' method must return a list or a string of args"
            )

        # Add the user options
        args_list = main_args + raw_args

        # Build the command line arguments
        cmdline = [self.executable]
        cmdline.extend(args_list)

        # Add persistent builder options
        cmdline.extend(self.persistant_options)

        # if warning_as_error is True, we check if the builder has registered
        # a command line option for that purpose. If it's the case, we check that
        # this option was not already define
        if warning_as_error is True:
            wae_args = self.get_warning_as_error_arg()
            if isinstance(wae_args, str):
                wae_args = wae_args.split()
            if not all(wae_arg in cmdline for wae_arg in wae_args):
                cmdline.extend(wae_args)

        # Create the build info object
        buildinfo = BuildInfo(
            app=app,
            builder=self.name,
            project_file=project_file,
            cmdline=cmdline,
            variant_args=variant_args,
        )
        self.display_build_info(buildinfo)

        # Run pre_build method if required
        self.pre_build(buildinfo, **variables)

        # Get an approximation build time execution
        time_started = time.time()

        # Build the application. In case the command is not found, we catch
        # the exception and make a result out of it.
        try:
            status_code, output = self.execute(cmdline, **kwargs)
        except BuildCommandNotFound as e:
            build_result = BuildResult(Result(Status.FAILURE, str(e)), str(e))
        else:
            # Get the execution time of the build
            execution_time = time.time() - time_started

            # Load the parser, parse and interpret the results
            parser = self.parser(warning_as_error)
            build_result = parser.execute(status_code, output)
            build_result.execution_time = execution_time

        # Save the build info for later use
        build_result.build_info = buildinfo

        # Display the result. Let the user re-define that if wanted
        self.display_result(build_result)

        # Save the log into the artifact path
        if output_file:
            with open(output_file, "w+") as f:
                f.write(build_result.output)

        # call the post_build method
        output_dir = Path(output_file).parent
        self.post_build(build_result, output_dir, **variables)

        return build_result

    @abstractmethod
    def get_main_args(
        self, project_file: Union[str, os.PathLike], **variant_args
    ) -> list[str]:
        """Return the main command line argument for the builder"""
        pass

    def execute(
        self, commands: list[str], **subprocess_args: Any
    ) -> Tuple[int, Union[str, bytes], str]:
        return self._execute(commands, **subprocess_args)

    def display_result(self, build_result: BuildResult) -> None:
        """Display build result"""
        status_msg = build_result.get_status_message()
        color = "green"
        if build_result.is_fail:
            color = "red"
        if build_result.result.message:
            terminal.line(build_result.result.message)
        terminal.line(f"Result: {status_msg}\n", fg=color)

    def pre_build(self, build_info: BuildInfo, **variables):
        """Method executed just before the build execution"""
        pass

    def post_build(self, result: BuildResult, output_dir: str, **variables):
        """Method executed after the build execution"""
        pass

    def display_build_info(self, buildinfo: BuildInfo) -> None:
        terminal.line(str(buildinfo))

    def _execute(
        self,
        command: Sequence[Any],
        silent: bool = True,
        shell: Union[None, bool] = None,
        env: Union[None, Mapping[str, str]] = None,
        **subprocess_kwargs: Any,
    ) -> Tuple[int, Union[str, bytes], str]:
        """
        Handles executing the command on the shell and consumes and returns
        the returned information (status_code, stdout/stderr)
        """

        # Output of the command. Represented as a list to append evert
        # line we read from the subprocess. This allow real-time output
        output = []

        # Don't automatically merge with os.environ for security reasons.
        # Make this forwarding explicit rather than implicit.
        inline_env = env
        env = os.environ.copy()
        if inline_env is not None:
            env.update(inline_env)

        cmd_not_found_exception = FileNotFoundError
        if os.name == "nt":
            cmd_not_found_exception = OSError

        # Make sure the command is a list if shell is False and the command
        # argument given is a string. Otherwise we will have a command error
        if isinstance(command, str):
            command = command.split(" ")

        logger.debug("Running following commands: {}\n".format(command))

        try:
            process = subprocess.Popen(
                command,
                shell=shell is not None and shell or self.use_shell,
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
                **subprocess_kwargs,
            )

            if silent is True:
                print("Building...")

            # Display realtime output and save it
            logger.debug("Command generated following output: ")
            while True:
                stdout = process.stdout.readline()
                return_code = process.poll()
                if not stdout and return_code is not None:
                    break
                if stdout:
                    stdout = safe_decode(stdout)
                    if silent is False:
                        sys.stdout.write(stdout)
                        sys.stdout.flush()
                    output.append(stdout)

        except cmd_not_found_exception as err:
            raise BuildCommandNotFound(command, err) from err
        else:
            output = "".join(output)
            return process.returncode, output

    def get_warning_as_error_arg(self) -> Union[str, list]:
        """Return the command line argument that enable warning as error"""
        return []



import codecs
import sys

from typing import Any, AnyStr, Optional, Union


def safe_decode(s: Union[AnyStr, None]) -> Optional[str]:
    """Safely decodes a binary string to unicode"""
    if isinstance(s, str):
        return s
    elif isinstance(s, bytes):
        return s.decode(sys.getfilesystemencoding(), "surrogateescape")
    elif s is None:
        return None
    else:
        raise TypeError("Expected bytes or text, but got %r" % (s,))


def to_text(obj: Any, encoding: str = "utf-8", nonstring="simplerepr"):
    """Make sure that a string is a text string

    :returns: Typically this returns a text string. If a nonstring object is
        passed in this may be a different type depending on the strategy
        specified by nonstring.  This will never return a byte string.

    """
    if isinstance(obj, str):
        return obj

    if isinstance(obj, bytes):
        try:
            codecs.lookup_error("surrogateescape")
            errors = "surrogateescape"
        except LookupError:
            errors = "strict"
        return obj.decode(encoding, errors=errors)

    if nonstring == "simplerepr":
        try:
            value = str(obj)
        except UnicodeError:
            try:
                value = repr(obj)
            except UnicodeError:
                return ""
    elif nonstring == "passthru":
        return obj

    return to_text(value, encoding)


def to_bytes(obj: Union[str, bytes]):
    """Make sure that a string is a byte string"""
    if isinstance(obj, bytes):
        return obj

    original_errors = None
    if isinstance(obj, str):
        try:
            # Try this first as it's the fastest
            return obj.encode("utf-8")
        except UnicodeEncodeError:
            if original_errors in (None, "surrogate_then_replace"):
                # We should only reach this if encoding was non-utf8 original_errors was
                # surrogate_then_escape and errors was surrogateescape

                # Slow but works
                return_string = obj.encode("utf-8", "surrogateescape")
                return_string = return_string.decode("utf-8", "replace")
                return return_string.encode("utf-8", "replace")
            raise

    try:
        value = str(obj)
    except UnicodeError:
        try:
            value = repr(obj)
        except UnicodeError:
            # Giving up
            return to_bytes("")

    return to_bytes(value)



from typing import Union

from socon.core.manager import Hook
from pydantic import BaseModel, ConfigDict

from socon_embedded.schema.task import Task


class ActionSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")


class ActionBase(Hook, abstract=True):

    manager = "ActionManager"
    schema = None

    def __init__(self, task: Task) -> None:
        self._task = task
        self.args = self.schema(**self._task.args)

    def run(self, tasks_vars=None):
        """Run the task"""
        return {"failed": False}

    def cleanup(self, force: bool = False):
        """Method to perform a clean up at the end of an action plugin execution

        Action plugins may override this if they deem necessary
        """



from socon_embedded.action import ActionBase, ActionSchema


class MoveActionShema(ActionSchema):
    pass


class MoveAction(ActionBase):
    name = "move"
    action_schema = MoveActionShema

    def run(self, tasks_vars=None):
        pass



from argparse import ArgumentParser
from pathlib import Path

from socon_embedded.executor.app_executor import AppRegistryExecutor
from socon_embedded.management.commands.build import BuildCommandInterface

from socon.core.management.base import Config
from socon.core.registry.config import ProjectConfig

from socon_embedded.managers import get_builder_manager


class BuildFromFile(BuildCommandInterface):
    name = "fromfile"

    def add_build_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "--file",
            help="Path to the YML application configuration file",
            required=True,
            type=Path,
        )

    def handle_build(
        self,
        config: Config,
        project_config: ProjectConfig,
        filters: dict = {},
        excludes: dict = {},
        variant_args_filters: dict = {},
        context: dict = {},
        exit_on_error: bool = False,
        warning_as_error: bool = False,
        artifact_dir: str = None,
    ) -> str:
        # Get all the registries that the user want to run
        app_registry = config.getoption("file")

        # Create the RegistryExecutor that will handle the execution of
        # all registries
        regexec = AppRegistryExecutor.from_file(
            app_registry,
            context=context,
            project_config=project_config,
            builder_manager=get_builder_manager(),
        )

        # Build the application using the selected registry
        regexec.build(
            filters=filters,
            excludes=excludes,
            variant_args_filters=variant_args_filters,
            exit_on_error=exit_on_error,
            warning_as_error=warning_as_error,
            output_dir=artifact_dir,
        )

        # Make a report at the root of the artifact directory
        regexec.create_report("results.xml", artifact_dir)



from unittest.mock import MagicMock

from socon_embedded.executor.task_result import TaskResult


class TestTaskResult:
    def test_task_result_basic(self):
        mock_task = MagicMock()
        _ = TaskResult(mock_task, dict())

    def test_task_result_is_failed(self):
        mock_task = MagicMock()

        # test with no failed in result
        tr = TaskResult(mock_task, dict())
        assert tr.is_failed() is False

        # test with failed in result
        tr = TaskResult(mock_task, dict(failed=True))
        assert tr.is_failed() is True



import os

from pathlib import Path
from typing import Any, Dict, Union

from jinja2 import Environment, FileSystemLoader


def jinja_resolve(file: Union[str, os.PathLike], context: Dict[str, Any]) -> str:
    """Resolve a file with the given context"""
    file = Path(file).expanduser().resolve()
    template_loader = FileSystemLoader(file.parent)
    env = Environment(loader=template_loader)
    template = env.get_template(file.name)
    return template.render(**context)



from socon_embedded.schema.task import Task

from socon_embedded.executor.task_executor import TaskExecutor


class BaseTestTask:

    def run_with_task_executor(self, task: Task) -> dict:
        executor = TaskExecutor(task)
        return executor.run()



import pytest

from socon.core.management import call_command


class TestBuildCommand:

    def test_check_available_subcommands(self, capsys):
        """Check that all subcommands are present"""
        with pytest.raises(SystemExit):
            call_command("build")
        captured = capsys.readouterr()
        assert "fromfile (P)" in captured.out



import codecs
import re

from ast import literal_eval

from socon_embedded.utils.converter import to_text

# Decode escapes adapted from rspeer's answer here:
# http://stackoverflow.com/questions/4020539/process-escape-sequences-in-a-string-in-python
_HEXCHAR = "[a-fA-F0-9]"
_ESCAPE_SEQUENCE_RE = re.compile(
    r"""
    ( \\U{0}           # 8-digit hex escapes
    | \\u{1}           # 4-digit hex escapes
    | \\x{2}           # 2-digit hex escapes
    | \\N\{{[^}}]+\}}  # Unicode characters by name
    | \\[\\'"abfnrtv]  # Single-character escapes
    )""".format(
        _HEXCHAR * 8, _HEXCHAR * 4, _HEXCHAR * 2
    ),
    re.UNICODE | re.VERBOSE,
)


def _decode_escapes(s: str) -> str:
    def decode_match(match):
        return codecs.decode(match.group(0), "unicode-escape")

    return _ESCAPE_SEQUENCE_RE.sub(decode_match, s)


def parse_key_value(args: str) -> dict:
    """Convert a string of key/value items to a dict"""
    args = to_text(args, nonstring="passthru")
    kwargs = {}

    if args is not None:
        splitted_args = split_args(args)
        for arg in splitted_args:
            arg = _decode_escapes(arg).split("=")
            key, value = arg[0], arg[1]

            # unquote the value if necessary
            if (
                len(value) > 1
                and value[0] == value[-1]
                and value[0] in ('"', "'")
                and value[-2] != "\\"
            ):
                value = value[1:-1]

            if value.startswith(("[", "{")):
                try:
                    evald = literal_eval(value)
                except (ValueError, SyntaxError):
                    continue
                if isinstance(evald, (dict, list)):
                    value = evald

            kwargs[key] = value

    return kwargs


def split_args(args: str) -> list:
    """Split args on whitespace and other elements"""
    if not args:
        return []

    # the list of params parsed out of the arg string
    # this is going to be the result value when we are done
    return [p for p in re.findall(r"(?:[^\s]+=.+?(?=\s+\w+=|$))", args, re.DOTALL)]



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    from socon_embedded.builder import BuildInfo


class Status:
    PASS: int = 0
    FAILURE: int = 1
    SKIPPED: int = -1


@dataclass
class Result:
    status_code: int = Status.PASS
    message: Optional[str] = None
    text: Optional[str] = None


class BuildResult:
    """Base class that stores build result"""

    def __init__(
        self,
        result: Result,
        output: Optional[str] = None
    ) -> None:
        self.result = result
        self.output = output
        self.build_info: BuildInfo = None
        self.execution_time: Union[float, int] = 0

    @property
    def is_skipped(self) -> bool:
        """Is the result is pass or fail"""
        return self.result.status_code <= Status.SKIPPED

    @property
    def is_fail(self) -> bool:
        """Is the result is pass or fail"""
        return self.result.status_code >= Status.FAILURE

    def get_status_message(self) -> str:
        if self.is_fail:
            return 'FAIL'
        elif self.is_skipped:
            return 'SKIPPED'
        return 'PASS'



from pathlib import Path

from socon_embedded.schema.task import Task
from tests.action import BaseTestTask


class TestCopyTask(BaseTestTask):

    def test_copy_file_to_a_folder(self, tmpdir):
        temp_d_1 = tmpdir.mkdir("sub_1").join("hello.txt")
        temp_d_1.write("content")
        temp_d_2 = tmpdir.mkdir("sub_2")
        task = Task(
            **{
                "name": "Copy file to a folder",
                "copy": {"src": str(temp_d_1), "dest": str(temp_d_2)},
            }
        )
        result = self.run_with_task_executor(task)
        assert result["failed"] is False
        assert Path(temp_d_2, "hello.txt").exists() == True

    def test_copy_folder_to_a_new_folder(self, tmpdir):
        temp_d_1 = tmpdir.mkdir("sub_1").join("content.txt")
        temp_d_1.write("content")
        destination = Path(tmpdir, "sub_2")
        task = Task(
            **{
                "name": "Copy folder to a folder",
                "copy": {"src": str(Path(tmpdir, "sub_1")), "dest": str(destination)},
            }
        )
        result = self.run_with_task_executor(task)
        assert result["failed"] is False
        assert Path(destination).exists() == True
        assert Path(destination, "content.txt").exists() == True

    def test_copy_content_to_a_file(self, tmpdir):
        temp_d_1 = tmpdir.mkdir("sub_1").join("hello.txt")
        task = Task(
            **{
                "name": "Copy content to a file",
                "copy": {"content": "Simple content", "dest": str(temp_d_1)},
            }
        )
        result = self.run_with_task_executor(task)
        assert result["failed"] is False
        assert temp_d_1.read() == "Simple content"



from _pytest.tmpdir import tmp_path_factory
from pydantic import ValidationError
import yaml

from pathlib import Path
from socon_embedded.exceptions import ParserError
from socon_embedded.schema.task import Task

import pytest


class TestTaskSchema:

    def _load_task(self, file: Path) -> Task:
        with open(file, "r") as fp:
            data = yaml.safe_load(fp)
        return Task(**data)

    def test_bad_action_module(self, datafix_dir):
        msg = "couldn't resolve module/action 'not_exist'"
        with pytest.raises(ParserError, match=msg):
            self._load_task(datafix_dir / "bad_action.yml")

    def test_no_action_module(self, datafix_dir):
        msg = "no module/action detected in task"
        with pytest.raises(ParserError, match=msg):
            self._load_task(datafix_dir / "no_action.yml")

    def test_with_action_module_set(self, datafix_dir):
        task = self._load_task(datafix_dir / "set_action.yml")
        assert task.action.name == "copy"

    def test_autodetect_action_with_args(self, datafix_dir):
        task = self._load_task(datafix_dir / "copy_schema.yml")
        assert task.action.name == "copy"
        assert task.args == {"src": "a", "dest": "b"}

    def test_mulitple_valid_action(self, datafix_dir):
        msg = "conflicting action statements"
        with pytest.raises(ParserError, match=msg):
            self._load_task(datafix_dir / "multiple_valid_action.yml")

    def test_check_action_schema_validation(self, datafix_dir):
        msg = "extra_forbidden"
        with pytest.raises(ValidationError, match=msg):
            self._load_task(datafix_dir / "action_extra_args.yml")


class TestCopyActionTask:

    def test_content_with_dest_as_dir(self, tmp_path):
        d = tmp_path / "sub"
        d.mkdir()
        msg = "Can not use 'content' with a dir as 'dest'"
        with pytest.raises(ValidationError, match=msg):
            Task(
                **{
                    "name": "test",
                    "copy": {"content": "File content", "dest": d.as_posix()},
                }
            )

    def test_mutual_exclusive_src_content(self):
        msg = "Only 'src' or 'content' must be supply at the same time"
        with pytest.raises(ValidationError, match=msg):
            Task(**{"name": "test", "copy": {"src": "", "content": "", "dest": ""}})

    def test_src_or_content_must_be_supply(self):
        msg = "'src' or 'content' field must be supply"
        with pytest.raises(ValidationError, match=msg):
            Task(**{"name": "test", "copy": {"dest": ""}})



from pathlib import Path

from socon_embedded.schema.task import Task

from tests.action import BaseTestTask


class TestCallCommandTask(BaseTestTask):

    def test_run_createcontainer_command(self, tmpdir):
        container = Path(tmpdir, "task_container")
        task = Task(
            **{
                "name": "Run simple command",
                "call_command": {
                    "cmd": "createcontainer",
                    "argv": ["task_container", "--target", str(container)],
                },
            }
        )
        result = self.run_with_task_executor(task)
        assert result["failed"] is False
        assert Path(tmpdir, "task_container").exists() is True



import os
import tempfile
import json
import shutil

from pathlib import Path
from typing import Optional

from socon_embedded.action import ActionBase, ActionSchema

from pydantic import model_validator

from socon_embedded.utils.converter import to_bytes


class CopySchema(ActionSchema):
    dest: str
    src: Optional[str] = None
    content: Optional[str] = None

    @model_validator(mode="after")
    def basic_validation(self):
        """If src is a directory, dest must be a directory too."""
        if self.src is None and self.content is None:
            raise ValueError("'src' or 'content' field must be supply")
        if self.src is not None and self.content is not None:
            raise ValueError("Only 'src' or 'content' must be supply at the same time")
        if self.content is not None and Path(self.dest).is_dir():
            raise ValueError("Can not use 'content' with a dir as 'dest'")
        return self


class CopyAction(ActionBase):
    name = "copy"
    schema = CopySchema

    def run(self, tasks_vars=None):
        if tasks_vars is None:
            tasks_vars = dict()

        # Result of the task stored in a dictionary
        result = super().run(tasks_vars)

        source = self.args.src
        content = self.args.content

        # Define content_tempfile in case we set it after finding content populated.
        content_tempfile = None

        if content is not None:
            try:
                # If content comes to us as a dict it should be decoded json.
                # We need to encode it back into a string to write it out.
                if isinstance(content, dict) or isinstance(content, list):
                    content_tempfile = self._create_content_tempfile(
                        json.dumps(content)
                    )
                else:
                    content_tempfile = self._create_content_tempfile(content)
                source = content_tempfile
            except Exception as err:
                result["failed"] = True
                result["msg"] = f"could not write content temp file: {err}"
                return result

        source = Path(source).expanduser().absolute()
        if source.is_dir():
            shutil.copytree(source, self.args.dest, dirs_exist_ok=True)
        else:
            shutil.copy2(source, self.args.dest)

        return result

    def _create_content_tempfile(self, content):
        """Create a tempfile containing defined content"""
        fd, content_tempfile = tempfile.mkstemp(prefix=".")
        f = os.fdopen(fd, "wb")
        content = to_bytes(content)
        try:
            f.write(content)
        except Exception as err:
            os.remove(content_tempfile)
            raise Exception(err)
        finally:
            f.close()
        return content_tempfile



from typing import Optional

from socon.core.management import call_command
from socon.core.management.base import CommandError
from socon_embedded.action import ActionBase, ActionSchema


class Schema(ActionSchema):
    cmd: str
    argv: Optional[list[str]] = []


class CallCommandAction(ActionBase):
    name = "call_command"
    schema = Schema

    def run(self, tasks_vars=None):
        result = super().run(tasks_vars)
        try:
            call_command(self.args.cmd, *self.args.argv)
        except CommandError as e:
            result["failed"] = True
            result["msg"] = f"Command error: {e}"
        return result



from typing import Any

from socon_embedded.schema.apps import AppConfig, AppRegistry


class TestSchemaCreation:

    def _validate(self, obj: Any):
        AppRegistry.model_validate(obj)

    def test_no_apps_registry(self):
        """Create a simple registry without any apps"""
        registry = AppRegistry(name="reg")
        assert registry.apps == []

    def test_add_app_to_registry(self):
        """Test that we can add an application to the registry after it's creation"""
        registry = AppRegistry(name="reg")
        registry.add_application("foo")
        assert len(registry.apps) == 1
        self._validate(registry)

    def test_add_group_to_an_existing_app(self):
        """
        Test that the same application will not be created twice and that we will
        only add the missing group
        """
        registry = AppRegistry(name="reg")
        app = registry.add_application("foo", group=1)
        assert app.group == [1]
        app = registry.add_application("foo", group=2)
        assert len(registry.apps) == 1
        assert app.group == [1, 2]
        self._validate(registry)

    def test_get_appication_in_registry(self):
        """Check the internal get application method"""
        app = AppConfig(name="foo")
        registry = AppRegistry(name="reg", apps=[app])
        assert registry._get_app("foo") == app

    def test_add_builder_to_app(self):
        """Check that we can add a builder to an application. Check also that
        creating twice the same builder will only add one (the first one)"""
        registry = AppRegistry(name="reg")
        app = registry.add_application("foo")
        app.add_builder("gcc", "path/to/file")
        assert len(app.builders) == 1

        # Do the same thing twice to check that we do not create two =
        # exact same builders
        app.add_builder("gcc", "path/to/file")
        assert len(app.builders) == 1

        # Vaidate the global model
        self._validate(registry)



from abc import ABC, abstractmethod

from socon_embedded.builder.result import BuildResult, Result, Status


class Parser(ABC):
    """Base class for all builder parsers"""

    def __init__(self, warning_as_error: False) -> None:
        self.warning_as_error = warning_as_error

    def execute(self, builder_status: int, output: str) -> BuildResult:
        """Parse and interpret the build result"""
        result = self.parse(builder_status, output)
        if not isinstance(result, Result):
            raise ValueError(
                "Parser class method parse(...) must return a Result based class"
            )
        return BuildResult(result, output)

    @abstractmethod
    def parse(self, builder_status, output: str) -> Result:
        """
        Parse the output of the builder. Method must return a Result instance
        or a Result based class (e.g Failure, SKipped).
        """
        pass


class DefaultParser(Parser):
    """
    Default Builder parser. This parser only parse for basic error and warning
    log. Return a dictionary with all error and warning texts.
    """

    def parse(self, builder_status: int, output: str) -> Result:
        if builder_status >= Status.FAILURE:
            return Result(
                builder_status, message="Error(s) found while building"
            )
        return Result(0, message="No error(s) found")



from socon_embedded.schema.task import Task


class TaskResult:
    """
    This class is responsible for interpreting the resulting data
    from an executed task, and provides helper methods for determining
    the result of a given task.
    """

    def __init__(self, task: Task, return_data: dict):
        self._task = task
        self._result = return_data

    @property
    def task_name(self):
        return self._task.name

    def is_changed(self):
        return self._check_key("changed")

    def is_skipped(self):
        return self._result.get("skipped", False)

    def is_failed(self):
        return self._check_key("failed")

    def _check_key(self, key):
        """get a specific key from the result or its items"""
        return self._result.get(key, False)



from __future__ import annotations

import pkgutil

from pathlib import Path

from socon.core.management.subcommand import SubcommandManager
from socon.core.manager import BaseManager, managers
from socon.core.registry.config import RegistryConfig
from socon.core.registry import registry, projects


# ------------------------- Build subcommand manager ------------------------- #


class SubcommandBuilderManager(SubcommandManager):
    name = "build_manager"
    lookup_module = SubcommandManager.lookup_module + ".build"

    def get_modules(self, config: type[RegistryConfig]) -> list:
        modules = super().get_modules(config)
        modules.extend(["socon_embedded.management.commands.subcommands.from_file"])
        return modules


# ------------------------------ Plugin manager ------------------------------ #


class AppManager(BaseManager):
    name = "app_manager"
    lookup_module = "apps"

    def get_modules(self, config: type[RegistryConfig]) -> list[str]:
        modules = super().get_modules(config)
        if isinstance(modules, str):
            modules = [modules]
        modules.extend(["socon_embedded.apps"])
        return modules


class BuilderManager(BaseManager):
    name = "builder"
    lookup_module = "builder"


def get_builder_manager():
    # Load the builder manager to get access to all registered builders
    # in the common config or in the current project
    builder_manager = managers.get_manager("builder")
    builder_manager.find_hooks_impl(registry.get_user_common_config())
    try:
        builder_manager.find_hooks_impl(projects.get_project_config_by_env())
    except LookupError:
        pass

    return builder_manager


def get_action_manager() -> ActionManager:
    # Load the action manager to get access to all registered actions
    # in the common config or in the current project
    manager = managers.get_manager("ActionManager")
    manager.find_all()
    return manager


class ActionManager(BaseManager):
    name = "ActionManager"
    lookup_module = "action"

    def get_modules(self, config: type[RegistryConfig]) -> list[str]:
        modules = super().get_modules(config)
        if isinstance(modules, str):
            modules = [modules]

        current_dir = Path(__file__).resolve().parent
        action_dir = Path(current_dir, "action")
        action_modules = [
            name
            for _, name, is_pkg in pkgutil.iter_modules([str(action_dir)])
            if not is_pkg and not name.startswith("_")
        ]

        # Loop over every module and get every class that are not abstract.
        # From these class we are going to get the commands that are
        # sublcalss of BaseCommand or ProjectCommand
        for action_module in action_modules:
            modules.append("socon_embedded.action.{}".format(action_module))
        return modules
